// Myrinet
//
// Guto, 05.05.1998, 15.05.1998

#ifndef __myrinet_h
#define __myrinet_h 1

#include <pcimap.h>
#include <dmabuf.h>
#include <string.h>
#include <netinet/in.h>

#define MYRINET_PACKET_TYPE	0x8888U

#define LANAI_VERSION		3		/* LANai 4.1 */
#define LANAI_TIMEOUT		3		/* 4s */
#define LANAI_DMA_BURST		0x04U		/* 8 words */
 
#define MYRINET_PAD0_BASE	0x00000000UL	/* 512 Kbytes */
#define MYRINET_PROM_BASE	0x00080000UL	/* 128 bytes */
#define MYRINET_PAD1_BASE	0x00080080UL	/* 256 Kbytes - 128 bytes */
#define MYRINET_REGS_BASE	0x000C0000UL	/* 256 bytes */
#define MYRINET_PAD2_BASE	0x000C0100UL	/* 128 Kbytes - 256 bytes */
#define MYRINET_FPGA_BASE	0x000E0000UL	/* 64 bytes */
#define MYRINET_PAD3_BASE	0x000E0040UL	/* 64 bytes */
#define MYRINET_SRAM_BASE	0x00100000UL	/* 1 Mbyte */

#define MYRINET_PAD0_SIZE	(MYRINET_PROM_BASE - MYRINET_PAD0_BASE)
#define MYRINET_PROM_SIZE	(MYRINET_PAD1_BASE - MYRINET_PROM_BASE)
#define MYRINET_PAD1_SIZE	(MYRINET_REGS_BASE - MYRINET_PAD1_BASE)
#define MYRINET_REGS_SIZE	(MYRINET_PAD2_BASE - MYRINET_REGS_BASE)
#define MYRINET_PAD2_SIZE	(MYRINET_FPGA_BASE - MYRINET_PAD2_BASE)
#define MYRINET_FPGA_SIZE	(MYRINET_PAD3_BASE - MYRINET_FPGA_BASE)
#define MYRINET_PAD3_SIZE	(MYRINET_SRAM_BASE - MYRINET_PAD3_BASE)
#define MYRINET_SRAM_SIZE	0x00100000UL
#define MYRINET_SIZE		(MYRINET_PAD0_SIZE + MYRINET_PROM_SIZE \
	 			+ MYRINET_PAD1_SIZE + MYRINET_REGS_SIZE \
				+ MYRINET_PAD2_SIZE + MYRINET_FPGA_SIZE \
				+ MYRINET_PAD3_SIZE + MYRINET_SRAM_SIZE )

// MYRINET special register bits
#ifndef NRES_ENABLE_BIT
#define NRES_ENABLE_BIT         0x00000001UL
#define CRC_ENABLE_BIT          0x00000002UL
#endif

// FPGA[0] (CONTROL) bits
#define LANAI_WAKE_ON           0x00000000UL
#define MYRINET_RESET_ON	0x00000004UL
#define MYRINET_RESET_OFF	0x00000008UL
#define LANAI_INT_DISABLE       0x00000010UL
#define LANAI_INT_ENABLE        0x00000020UL
#define LANAI_RESET_ON          0x00000040UL
#define LANAI_RESET_OFF         0x00000080UL

// Interrupt Status Register bits
#define DBG_BIT         	0x80000000
#define HOST_SIG_BIT    	0x40000000
#define LAN7_SIG_BIT    	0x00800000
#define LAN6_SIG_BIT    	0x00400000
#define LAN5_SIG_BIT    	0x00200000
#define LAN4_SIG_BIT    	0x00100000
#define LAN3_SIG_BIT    	0x00080000
#define LAN2_SIG_BIT    	0x00040000
#define LAN1_SIG_BIT    	0x00020000
#define LAN0_SIG_BIT    	0x00010000
#define WORD_RDY_BIT    	0x00008000
#define HALF_RDY_BIT    	0x00004000
#define SEND_RDY_BIT    	0x00002000
#define LINK_INT_BIT    	0x00001000
#define NRES_INT_BIT    	0x00000800
#define WAKE_INT_BIT    	0x00000400
#define OFF_BY_2_BIT   	 	0x00000200
#define OFF_BY_1_BIT    	0x00000100
#define TAIL_INT_BIT    	0x00000080
#define WDOG_INT_BIT    	0x00000040
#define TIME_INT_BIT    	0x00000020
#define DMA_INT_BIT     	0x00000010
#define SEND_INT_BIT    	0x00000008
#define BUFF_INT_BIT    	0x00000004
#define RECV_INT_BIT    	0x00000002
#define BYTE_RDY_BIT    	0x00000001
#define ORUN2_BIT       	OFF_BY_2_BIT
#define ORUN1_BIT       	OFF_BY_1_BIT
#define ORUN2_INT_BIT   	OFF_BY_2_BIT
#define ORUN1_INT_BIT   	OFF_BY_1_BIT
 
struct Myrinet_PROM
{
    unsigned long  lanai_clockval;
    unsigned short lanai_cpu_version;
    unsigned char  lanai_board_id[6];
    unsigned long  lanai_sram_size;
    unsigned char  fpga_version[32];
    unsigned char  more_version[16];
    unsigned short delay_line_value;
    unsigned short board_type;
    unsigned short bus_type;
    unsigned short product_code;
    unsigned long  serial_number;
    unsigned short reserved[26];
};

struct Myrinet_Registers
{
    volatile unsigned long ipf0;     /* context-0 state registers */
    volatile unsigned long cur0;
    volatile unsigned long prev0;
    volatile unsigned long data0;
    volatile unsigned long dpf0;
    volatile unsigned long ipf1;     /* context-1 state registers */
    volatile unsigned long cur1;
    volatile unsigned long prev1;
    volatile unsigned long data1;
    volatile unsigned long dpf1;
    volatile unsigned long isr;      /* interrupt status register */
    volatile unsigned long eimr;     /* external-longerrupt mask register */
    volatile unsigned long it;       /* longerrupt timer */
    volatile unsigned long rtc;      /* real-time clock */
    volatile unsigned long cks;      /* checksum */
    volatile unsigned long ear;      /* SBus-DMA exteral address */
    volatile unsigned long lar;      /* SBus-DMA local address */
    volatile unsigned long dma_ctr;  /* SBus-DMA counter */
    volatile unsigned long rmp;      /* receive-DMA polonger */
    volatile unsigned long rml;      /* receive-DMA limit */
    volatile unsigned long smp;      /* send-DMA polonger */
    volatile unsigned long sml;      /* send-DMA limit */
    volatile unsigned long smlt;     /* send-DMA limit with tail */
    unsigned long skip_0x5c;         /* skipped one word */
    volatile unsigned char rb;       /* receive byte */
    unsigned char skip_0x61;
    unsigned char skip_0x62;
    unsigned char skip_0x63;
    volatile unsigned short rh;      /* receive half-word */
    unsigned char skip_0x66;
    unsigned char skip_0x67;
    volatile unsigned long rw;       /* receive word */
    volatile unsigned long sa;       /* send align */
    volatile unsigned long sb;       /* single-send commands */
    volatile unsigned long sh;
    volatile unsigned long sw;
    volatile unsigned long st;
    volatile unsigned long dma_dir;  /* SBus-DMA direction */
    volatile unsigned long dma_sts;  /* SBus-DMA modes */
    volatile unsigned long timeout;
    volatile unsigned long myrinet;
    volatile unsigned long hw_debug; /* hardware debugging */
    volatile unsigned long led;      /* LED pin(s) */
    volatile unsigned long version;  /* the ex-window-pins register */
    volatile unsigned long activate; /* activate myrinet-link *//* 0x9C */
    unsigned long pad_a[(0xfc - 0xa0) / sizeof(long)];
    volatile unsigned long clock;  /* clock register 0xFC */
};


typedef volatile unsigned int Myrinet_FPGA[MYRINET_FPGA_SIZE / sizeof(int)];

typedef unsigned char Myrinet_SRAM[MYRINET_SRAM_SIZE];

class Myrinet_NIC {
public:
    Myrinet_NIC(){};

    void int_enable(){fpga[0] = htonl(LANAI_INT_ENABLE);}
    void int_disable(){fpga[0] = htonl(LANAI_INT_DISABLE);}
    void suspend() {
	fpga[0] = htonl(LANAI_RESET_ON);
	fpga[0] = htonl(MYRINET_RESET_ON);
    }
    void resume() {
	fpga[0] = htonl(MYRINET_RESET_OFF);
	fpga[0] = htonl(LANAI_RESET_OFF);
    }
    void reset() {
	fpga[0] = htonl(LANAI_WAKE_ON);

	int_disable();
	suspend();

        memset(sram, 0, MYRINET_SRAM_SIZE);
	unsigned int delay = ntohs(prom.delay_line_value);
	if(delay >= 0xffff)
	    delay = 0xe;

	fpga[0] = 0;
	fpga[0] = 0x4;
	for(int i = 8; i--; )
	    if((1 << i) & delay) {
		fpga[0] = 0x5;
		fpga[0] = 0x7;
		fpga[0] = 0x5;;
	    } else {
		fpga[0] = 0x4;
		fpga[0] = 0x6;
		fpga[0] = 0x4;
	    }
	fpga[0] = 0;                        

	regs.clock = prom.lanai_clockval;
	regs.eimr = 0;
	regs.myrinet = htonl(CRC_ENABLE_BIT | NRES_ENABLE_BIT);
	regs.version = htonl(LANAI_VERSION);
	regs.timeout = htonl(LANAI_TIMEOUT);

	regs.eimr = 0;
	regs.activate = htonl(1);
	regs.dma_sts = htonl(LANAI_DMA_BURST);

	resume();
	int_enable();
    };

public:
    unsigned char      	pad0[MYRINET_PAD0_SIZE];
    Myrinet_PROM       	prom;
    unsigned char      	pad1[MYRINET_PAD1_SIZE];
    Myrinet_Registers	regs;
    unsigned char      	pad2[MYRINET_PAD2_SIZE];
    Myrinet_FPGA      	fpga;
    unsigned char      	pad3[MYRINET_PAD3_SIZE];
    Myrinet_SRAM      	sram;
};

class Myrinet
{
protected:

public:
    Myrinet(int unit = 0);

    ~Myrinet();

    unsigned int align_length(unsigned int len)
	{return (len + sizeof(int) - 1) & ~(sizeof(int) - 1);}

    bool length_ok(unsigned int len)
	{return !(!len || (len & sizeof(int) - 1) || (len > mtu));}
    bool reset_ok();
    bool send_ok(void *buf, unsigned int *len);
    bool receive_ok(void *buf, unsigned int *len);


    int reset();
    int send(void *buf, unsigned int len);
    int receive(void *buf, unsigned int *len);


protected:
    void drop();

protected:
    PCI_Mapped_Device *dev;
    Myrinet_NIC *nic;
    DMA_Buffer *dmabuf;
    unsigned int mtu;
};


#endif // __myrinet_h
